/* Geiger Shield Sketch (Ver. 4.0c)                                     bHogan 12/11/11
 * This sketch is run on the Geiger Kit when the Geiger Shield is attached. It . . .
 * - outputs CPM & ~uSv/hr to the display and serial - like the default sketch.
 * - logs this info to an SD card with a timestamp from either the RTC or GPS
 * - if the GPS is used it also logs latitude & longitude
 * - uses an IR receiver to accept commands from a TV remote (Sony / RC5 protocol) 
 * - reads a set of 4 DIP switches used as follows:
 *    - #1 (D4): starts "dose mode" - see below
 *    - #2 (D5): turns off backlight - (v.4c - see below)
 *    - #3 (D6): reserved for user
 *    - #4 (D7): reserved for user
 *
 * Ver 2 - improves display in "dose mode"
 *       - creates enough SRAM for GPS by not allowing menu in that mode (use PC Setup sketch)
 *       - moves the EEPROM storage for the GMT Time Zone parameter.
 * Ver 3 - uses PROGMEM to reduce SRAM and reestablish menus in GPS mode
 * Ver 3b- provides a more realistic estimate of displayed counts using a moving average.
 *       - See comments in Default Sketch V6b
 *       - Fixed bug in logging count
 * Ver 4 - Updated with changes from Ver. 7 of the Default Geiger sketch.
 *       - DIP switch #1, or the menu, turns on "dose display" in this version
 *       - See comments in Default Sketch V7 * "Jumpers & Switches" page on the site.
 * Ver 4d- Addition of some nice enhancements with thanks to Gerald.  
 *       1. A IR_RC5 #define to use the Phillips RC5 IR protocol insted of the Sony protocol.
 *       2. A GPS_SKYTRAQ #define to use the SkyTraq GPS module instead of the EM-406A module.
 *       -  More info on the SkyTraq is at http://www.mr-lee-catcam.de/pe_cc_i6.htm
 *       3. The INFO key on the remote toggles the backlight on and off
 *          (DIP sw #2 also turns it on and off.)
 *       4. New menu option to toggle the piezo on and off. (This requires a hardware mod:
 *       - on the Geiger borard disconnect R13 from Vcc and reconnected it on D16.)
 *
 * PIN USAGE:
 * D2 = Interrupt from Geiger         D3 = Interrupt from IR receiver
 * D4, D5, D6, D7 = 4 user DIP switches
 * D10 (CS), D11 (MOSI), D12 (MISO), D13 (CLK) - for SD card 
 * A4 (SDA), A5 (SCL) - for IC2 display and RTC
 * (display is any 2x16 Hitachi HD44780 compatible LCD)
 
 * This program is free software; you can redistribute it and/or modify it under the
 * terms of the GNU General Public License as published by the Free Software
 * Foundation; either version 2.1 of the License, or any later version.
 * This program is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
 * PARTICULAR PURPOSE.  See the GNU General Public License for more details.
 * 
 * THIS PROGRAM AND IT'S MEASUREMENTS IS NOT INTENDED TO GUIDE ACTIONS TO TAKE, OR NOT
 * TO TAKE REGARDING EXPOSURE TO RADIATION. THE GEIGER KIT AND IT'S SOFTWARE ARE FOR
 * EDUCATIONAL PURPOSES ONLY. DO NOT RELY ON THIS PROGRAM IN HAZARDOUS SITUATIONS!
 *
 * NOTES:
 * - counts accumulate while in menu - by design
 * - RTC current RAM used / SRAM avail: 28114 / 466
 * - GPS current RAM used / SRAM avail: 30136 / 378 
 * TODO:
 * - menu system could work better
 * - use auto DST for RTC in addition to GPS
 * - If day & month & hours are 2 digits - date time string is 17 char & missing last sec.
 * - use another dip switch pin to control backlight
 */

// BEGIN USER PARAMETER AREA . . .
#define DEBUG       false               // if true shows available memory
#define USER_INFO   false               // displays and logs the user name set in shield.h
#define GPS_USED    false               // if true GPS used instead of RTC
#define GPS_SKYTRAQ false               // for EM-406A set to false
#define IR_RC5      false               // use Phillips RC5 IR protocol instead of Sony 

// date and time formatting options - apply to both the display and the log. 
#define DATE_FMT         DDMMYY         // pick date format - DDMMYY, MMDDYY, YYMMDD
#define SPEED_KMPH       true           // speed in km/hr if true - else MPH
#define AM_PM_FMT        false          // 12H format - if true must not disp seconds if used
#define NO_SECONDS       false          // set to true if seconds are not displayed or logged
#define AUTO_DST         true           // adjust for US DST if true & GPS used
// END USER PARAMETER AREA


#include <SD.h>                         // for SD Card
#include <Wire.h>                       // for RTC
#include <LiquidCrystal_I2C.h>          // I2C LCD display driver
#include "shield.h"                     // .h tab
#if (GPS_USED)
#include <TinyGPS.h>                    // MODIFIED LIB - parses GPS sentences
// MODIFIED LIB - _GPS_NO_STATS uncommented (and resulting compile error fixed) saves 80 bytes
#else
#include <DS1307RTC.h>                  // returns time as a time_t from RTC
#endif
#include <Time.h>                       // time functions
#include <EEPROM.h>                     // ATmega328's EEPROM used to store settings
#include <avr/pgmspace.h>               // for using PROGMEM

// instantiate the library and pass pins for (RS, Enable, D4, D5, D6, D7)
LiquidCrystal_I2C lcd(GPIO_ADDR,16,2);  // set the LCD address, 16 chars, 2 lines
#if (GPS_USED)
TinyGPS gps;                            // GPS conversion library
#endif
File logfile;                           // the logging file


void setup(){
  lcd.init();                           // initialize the lcd 
  lcd.backlight();                      // turn on the backlight
  Serial.begin(GPSRATE);                // comspec for GPS is 48,N,8,1
  attachInterrupt(0,GetEvent,FALLING);  // Geiger event on D2 triggers interrupt
  attachInterrupt(1,IR_ISR,FALLING);    // Catch IR sensor on D3 with interrupt 
//  digitalWrite(2, HIGH);              // set pull up on INT0
  pinMode(LED_PIN, OUTPUT);             // setup LED pin
  pinMode(ALARM_PIN, OUTPUT);           // setup Alarm pin
  pinMode(SWITCH_1,INPUT);              // setup user switches
  digitalWrite(SWITCH_1, HIGH);         // set 20K pullup on each switch (low active)
  pinMode(SWITCH_2,INPUT);
  digitalWrite(SWITCH_2, HIGH);
  pinMode(SWITCH_3,INPUT);
  digitalWrite(SWITCH_3, HIGH);
  pinMode(SWITCH_4,INPUT);
  digitalWrite(SWITCH_4, HIGH);
  pinMode(BUZZER_PIN, OUTPUT);          // sw control buzzer
  digitalWrite(BUZZER_PIN, HIGH);       // buzzer = on

  Get_Settings();                       // get the settings stored in EEPROM
  Blink(LED_PIN,3);                     // show it's alive (LED is on pin 14)

  lcd.setCursor(0,0);                   // HELPS WITH OLED DISPLAY
  clearDisp();                          // clear the screen
  // Using PROGMEM to display lits to save SRAM - strcpy_P loads lit in dispBuff
  strcpy_P(dispBuff, (char*)pgm_read_word(&(lit_table[0]))); //("  GEIGER LOGGER")
  lcd.print(dispBuff);
  lcd.setCursor(0,1);                   // set cursor on line 2
  strcpy_P(dispBuff, (char*)pgm_read_word(&(lit_table[7]))); //("  Version: ")
  lcd.print(dispBuff);
  delay (2000);                         // leave the banner up 2 sec.
  clearDisp();                          // clear the screen
#if (USER_INFO)                         // display owner if defined
  strcpy_P(dispBuff, (char*)pgm_read_word(&(lit_table[8]))); //("Owner")
  lcd.print(dispBuff);
  lcd.setCursor(0,1);                   // set cursor on line 2
  strcpy_P(dispBuff, (char*)pgm_read_word(&(lit_table[9]))); //("  (line 2)")
  lcd.print(dispBuff);
  delay (3000);                         // leave the banner up 3 sec.
  clearDisp();                          // clear the screen
#endif
  lcd.print(uSvRate,0);                 // display ratio used 
  strcpy_P(dispBuff, (char*)pgm_read_word(&(lit_table[1]))); //(" CPM to uSv")
  lcd.print(dispBuff);
  lcd.setCursor(0,1);                   // set cursor on line 2
  Vcc_mV = readVcc();                   // read Vcc voltage
  strcpy_P(dispBuff, (char*)pgm_read_word(&(lit_table[2]))); //("Running at ")
  lcd.print(dispBuff);
  lcd.print(Vcc_mV/1000. ,2);           // convert to Float, divide, and print 2 dec places
  lcd.print("V");
  delay (2500);                         // leave info up for 2.5 sec.
  clearDisp();                          // clear the screen
  initCard();                           // init the SD card
#if (GPS_USED)
  GPSinit();                            // init the GPS to send output only when requested
  setSyncProvider(gpsTimeSync);         // define the function that syncs the time with GPS
#else
  setSyncProvider(RTC.get);             // define the function that syncs the time with RTC
#endif

  clearDisp();                          // clear the screen
#if (DEBUG)                             // show available SRAM if DEBUG true
  strcpy_P(dispBuff, (char*)pgm_read_word(&(lit_table[3]))); //("RAM Avail: ")
  lcd.print(dispBuff);
  lcd.print(AvailRam());
  delay (2000);
  clearDisp();
#endif
  lcd.print("CPM ?");                   // display static "CPM"
  dispDateTime();                       // display date & time on second line
  dispPeriodStart = millis();           // start timing display CPM
  logPeriodStart = dispPeriodStart;     // start logging timer
  runCountStart = dispPeriodStart;      // start alternate timer
  dispCnt = 0;                          // start with fresh totals
  logCnt= 0;
  runCnt = 0;
}


void loop(){
  //uncomment 4 lines below for self check (1/166ms X 60 = 360 CPM
  //dispCnt++;
  //logCnt++;
  //runCnt++;
  //delay(167);                         // 167 mS ~= 6 Hz    

  Check_IR();                           // check if IR received a command

  checkSwitches();                      // check all 4 DIP switches
  if (SW1_On || doseParam){             // start dose display mode if SW 1 on or param set
    if (!doseMode)altDispCnt = 4;       // show dose display immediately
    doseMode = true;                    // set flag
  }
  else doseMode = false;

  if (Backlight){                        // check switch only if IR command did not turn backlight off
    if (SW2_On)lcd.noBacklight();          // if SW 2 on or param set, turn off backlight
    else lcd.backlight();
  }
  
  if (!inMenu){                         // none of this happens if you are in the menu
    if (millis() >= runCountStart + 10000){ // Collect 1 & 10 min. running counts every 10 sec.
      RunningCounts(runCnt);            // add counts
      runCnt = 0;                       // reset counts
      runCountStart = millis();         // reset the period time
    }

    if (millis() >= dispPeriodStart + dispPeriod){ // DISPLAY PERIOD
      DispCounts(dispCnt);                  // period is over - display counts
      if (!DoseDispOn) dispDateTime();      // display date/time on line 2 unless dose is displaying
      if (DoseDispOn) DoseDispOn = false;   // reset flag for next time
      dispCnt = 0;                          // reset counter unless continuous count
      dispPeriodStart = millis();           // reset the period time
    }

    if (millis() >= logPeriodStart + LoggingPeriod){ // LOGGING PERIOD
      if (SD_OK) LogCount(logCnt);      // pass in the counts to be logged
      logCnt = 0;                       // reset log event counter
      dispCnt = 0;                      // reset display event counter too
      dispPeriodStart = millis();       // reset display time too
      logPeriodStart = millis();        // reset log time
    }
  }
}


void DispCounts(unsigned long dcnt){    // calc and display predicted CPM & uSv/h
  //dispCPM = (dcnt * 60000) /(DISP_PERIOD);     // older simple calc method - now a moving average
  byte maxSamples = (60000 / DISP_PERIOD) / 2;   // number of sample periods in 30 seconds                     
  sampleCnt++;                                   // inc sample count - must be at least 1
  avgCnt += (dcnt - avgCnt) / sampleCnt;         // CALCULATE AVERAGE COUNT - moving average
  dispCPM = (avgCnt * 60000.0) / DISP_PERIOD;    // convert to CPM
  //handle reset of sample count - sample is for 1/2 min and reset. Options for reset value are:
  // "0" - throw away last average, "1" - keeps last average, "maxSamples -1" - keeps running avg.
  if (sampleCnt >= maxSamples) sampleCnt = 0;    // start a fresh average every 30 sec.
  // the following line gives a faster response when counts increase or decrease rapidly 
  if ((dcnt - avgCnt) > 9 || (avgCnt - dcnt) > 9) sampleCnt = 0;
  uSv = dispCPM / uSvRate;                       // make uSV conversion

  altDispCnt ++;                        // decide what screen will show . . .
  if (altDispCnt >=4){                  // the dose count screen will display every 4th display
    altDispCnt = 0;                     // reset the count
    if (doseMode && dispOneMin){        // don't disply unless there is something to display
      DispRunCounts();                  // display the dose mode screen
      return;                           // skip the normal count screen
    }
  }

  // display the counts on the LCD
  clearDisp();                          // clear the screen
  clearArea (0,0,16);                   // clear line 1

  lcd.print("CPM ");                    // display static "CPM"
  lcd.print(dispCPM);                   // display CPM on line 1
  lcd.setCursor(13 - getLength(uSv),0); // right justify the uSv!
  lcd.print(uSv,2);                     // display uSv/hr on line 1

  if (lowVcc) {                         // overwrite uSv if battery voltage low
    clearArea (9,0,8);
    lcd.print("Vcc=");
    lcd.print(Vcc_mV/1000. ,2);         //convert to Float, divide, and print 2 dec places
  }

  if (dispCPM > AlarmPoint && AlarmPoint > 0) {  // Alarm takes priority over low Vcc
    clearArea (11,0,6);
    lcd.print("ALARM");
    AlarmOn = true;
    digitalWrite(ALARM_PIN, HIGH);      // turn on alarm (set alarm pin to Vcc) 
  }
  else{
    AlarmOn = false;
    digitalWrite(ALARM_PIN, LOW);       // turn off alarm (set alarm pin to Gnd)
  }
}


void DispRunCounts(){ // create the screen that shows the running counts
  DoseDispOn = true;                    // set the flag
  clearDisp();                          // clear the screen
  lcd.print(" 1M ");                    // display 1 & 10 min lits
  lcd.setCursor(0,1);
  lcd.print("10M ");

  tempSum = 0;
  for (int i = 0; i <= ONE_MIN_MAX; i++){ // sum up 1 minute counts
    tempSum = tempSum + oneMinute[i];
  }
  temp_uSv = tempSum / uSvRate;         // calc uSv/h
  lcd.setCursor(9, 0);
  lcd.print("/");
  lcd.setCursor(4, 0);
  lcd.print(tempSum,DEC);               // display 1 minute CPM
  lcd.setCursor(13 - getLength(temp_uSv), 0); // right justify the uSv!
  lcd.print(temp_uSv,2);                // display 1 minute uSv

  if (dispTenMin) {                     // see if 10 minute average is available
    tempSum = 0;
    for (int i = 0; i <= TEN_MIN_MAX; i++){ // sum up 10 minute counts
      tempSum = tempSum + tenMinute[i];
    }
    tempSum = tempSum / 10;             // sum over 10 minutes so divide by that
    temp_uSv = tempSum / uSvRate;
    lcd.setCursor(9, 1);                // display first so it can be overwritten if needed
    lcd.print("/");
    lcd.setCursor(4, 1);
    lcd.print(tempSum,DEC);             // display 10 minute CPM
    lcd.setCursor(13 - getLength(temp_uSv), 1);  // right justify
    lcd.print(temp_uSv,2);              // display 10 minute uSv
  }
  else { // 10 minute average not available, so display seconds left to wait
    lcd.setCursor(3, 1);
    lcd.print(" wait "); 
    lcd.print(600 - (10*tenMinuteIndex),DEC); // show seconds left
    lcd.print(" sec"); 
  }
  // give the display a little more time for fast display periods
  if (DISP_PERIOD < 5000) delay(1500);  // a delay is OK here - interrupts are still handled
}


void RunningCounts(unsigned long dcnt){ // add to running counts and reset counters as needed
  // credit and thanks to Marek Schmidt for the routine and concept.
  // Add CPM of finished period to 1M and 10M array
  oneMinute[oneMinuteIndex] = dcnt;
  if(oneMinuteIndex >= ONE_MIN_MAX) {
    oneMinuteIndex = 0;
    dispOneMin = true;                  // indicate 1M average is available
  }
  else oneMinuteIndex++;

  tenMinute[tenMinuteIndex] = dcnt;
  if(tenMinuteIndex >= TEN_MIN_MAX) {
    tenMinuteIndex = 0;
    dispTenMin = true;                  // indicate 10M average is available
  }
  else tenMinuteIndex++;
  dcnt = 0;
}


void clearArea (byte col, byte line, byte nspaces){
  // starting at col & line, prints n spaces then resets the cursor to the start
  lcd.setCursor(col,line);
  for (byte i=0; i<nspaces; i++){
    lcd.print(" ");
  }
  lcd.setCursor(col,line);
}


void clearDisp (){
  // The OLED display does not always reset the cursor after a clear(), so it's done here
  lcd.clear();                          // clear the screen
  lcd.setCursor(0,0);                   // reset the cursor for the poor OLED
  lcd.setCursor(0,0);                   // do it again for the OLED
}


void Blink(byte led, byte times){ // just to flash the LED
  for (byte i=0; i< times; i++){
    digitalWrite(led,HIGH);
    delay (120);
    digitalWrite(led,LOW);
    delay (90);
  }
}


long readVcc() { // SecretVoltmeter from TinkerIt
  long result;
  // Read 1.1V reference against AVcc
  ADMUX = _BV(REFS0) | _BV(MUX3) | _BV(MUX2) | _BV(MUX1);
  delay(2); // Wait for Vref to settle
  ADCSRA |= _BV(ADSC); // Convert
  while (bit_is_set(ADCSRA,ADSC));
  result = ADCL;
  result |= ADCH<<8;
  result = 1126400L / result; // Back-calculate AVcc in mV
  return result;
}

void checkSwitches(){ // test all four user switches and set a flag true if ON
  SW1_On = ! digitalRead(SWITCH_1);
  SW2_On = ! digitalRead(SWITCH_2);
  SW3_On = ! digitalRead(SWITCH_3);
  SW4_On = ! digitalRead(SWITCH_4);  
}

int AvailRam(){ 
  int memSize = 2048;                   // if ATMega328
  byte *buf;
  while ((buf = (byte *) malloc(--memSize)) == NULL);
  free(buf);
  return memSize;
} 

byte getLength(unsigned long number){
  byte length = 0;

  for (byte i = 1; i < 10; i++){
    if (number > pow(10,i)) length = i;
    else return length +1;
  }
}

void GetEvent(){   // ISR triggered for each new event (count)
  dispCnt++;
  logCnt++;
  runCnt++;
}

